#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 12 02:57:59 2020

@author: eduardo
"""

import warnings

from .functions import set_dir_struct, riffle, file_names, set_directory, distribute_among_walkers
from .dataset import read_dataset
from .country import country
from .models import *




# neglecting warnings
warnings.filterwarnings('ignore')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

Description
----------
The country object contains information and actions 
related the countries that are important to understand 
the Covid dissemination

Informations
----------
    Author: Eduardo M.  de Morais
    Maintainer:
    Email: emdemor415@gmail.com
    Copyright:
    Credits:
    License:
    Version:
    Status: in development
    
"""

import pandas as pd
import numpy as np

from .functions import set_dir_struct, riffle, file_names, set_directory, distribute_among_walkers
from os import path

# empty dataframe to use in case of error
__EMPTY_DATAFRAME__ = pd.DataFrame({'ID' :[],
                                  'Updated' :[],
                                  'Confirmed' :[],
                                  'ConfirmedChange' :[],
                                  'Deaths' :[],
                                  'DeathsChange' :[],
                                  'Recovered' :[],
                                  'RecoveredChange' :[],
                                  'Latitude' :[],
                                  'Longitude' :[],
                                  'ISO2' :[],
                                  'ISO3' :[],
                                  'Country_Region' :[],
                                  'AdminRegion1' :[],
                                  'AdminRegion2' :[]})



# global variables to use as standard 
__DATASET_URL__ = 'https://raw.githubusercontent.com/microsoft/Bing-COVID-19-Data/master/data/Bing-COVID19-Data.csv'

__DATASET_FILENAME__ = 'Bing-COVID19-Data.csv'





def read_dataset(update_data = False,
                 url         = __DATASET_URL__,
                 local_dataset_filename = __DATASET_FILENAME__):
    '''
    Description
    ----------
    This function imports the Microsoft dataset from local source. If
    required, downloads the updated date from online source and saves 
    it on the tables directory, owerwritten the previous file.

    After import, converts the columns to the right types, change the
    name of some countries and creates a new column for active infections.

    Arguments
    ----------
    update_data: bool
        Controls if the data will be updated

    url: str
        URL to download the online sourde

    local_dataset_filename: str
        Name (without path) of localbase dataset

    Return
    ----------
    df: Pandas.DataFrame
        Final dataset
        
    '''

    # join filename with directory
    #filename = path.join(tables_directory, local_dataset_filename)
    filename = local_dataset_filename

    # updating when required
    if update_data:
        update_local_base(url,filename)
    
    #importing dataset
    df = import_from_localbase(filename)[0]

    # Filling NA values
    df[['ISO2','ISO3','Country_Region','AdminRegion1','AdminRegion2']] = df[['ISO2','ISO3','Country_Region','AdminRegion1','AdminRegion2']].fillna('')
    df[['Confirmed','ConfirmedChange','Deaths','DeathsChange','Recovered','RecoveredChange']] = df[['Confirmed','ConfirmedChange','Deaths','DeathsChange','Recovered','RecoveredChange']].fillna(0)
    
    # Type coversion
    df[['Confirmed','ConfirmedChange','Deaths','DeathsChange','Recovered','RecoveredChange']] = df[['Confirmed','ConfirmedChange','Deaths','DeathsChange','Recovered','RecoveredChange']].astype(float)
    df[['ISO2','ISO3','Country_Region','AdminRegion1','AdminRegion2']] = df[['ISO2','ISO3','Country_Region','AdminRegion1','AdminRegion2']].astype(str)
    
    # Replacing Countries names
    df['Country_Region'] = df['Country_Region'].replace('China (mainland)','China')
    
    # Creating a new column
    df['Actives'] = df['Confirmed']-df['Deaths']-df['Recovered']

    return df


def import_from_url(url):
    '''
    Description
    ----------
    
    Arguments
    ----------
        
    '''

    try:
        print('[status]: Downloading dataset from source.')
        dataframe = pd.read_csv(url,sep=",",parse_dates=['Updated'])
        return [dataframe,True]

    except:
        # When occours some
        print('[error]: It was not possible to download data from online source.')
        return [__EMPTY_DATAFRAME__,False]



def import_from_localbase(filename):
    '''
    Description
    ----------
    
    Arguments
    ----------
        
    '''
    try:
        if not path.exists(filename):
            print('[warng]: There is no previous dataset on the localfile. Dataset must be downloaded from online source.')
            print('[mssge]: Getting data from online source :'+__DATASET_URL__+'.')
            update_local_base(__DATASET_URL__,filename)

        print('[status]: Importing dataset from localbase.')
        dataframe = pd.read_csv(filename,sep=",",parse_dates=['Updated'])
        return [dataframe,True]

    except:
        print('[error]: There is no local base file.')
        return [__EMPTY_DATAFRAME__,False]


def export_to_local_base(df,address):
    '''
    Description
    ----------
    
    Arguments
    ----------
        
    '''
    try:
        print('[status]: Updating local dataset base.')
        df.to_csv(address,sep=',',index=False)
    except:
        print('[error]: It was not possible to export data to local base.')


def update_local_base(url,address):
    '''
    Description
    ----------
    
    Arguments
    ----------
        
    '''

    try:
        df,__IMPORT_STATUS__ = import_from_url(url)

        if __IMPORT_STATUS__ :
            export_to_local_base(df,address)
        else:
            if not path.exists(address):
                print('[error]: Is impossible to access the data from local base or online source.')
            else:
                print('[mssge]: Keeping the previous data on local base.')
    except:
        print('[error]: It was not possible update local base.')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Description
----------
This module gives functions and classes to manipulate the global statistical
properties of the model

Informations
----------
    Author: Eduardo M.  de Morais
    Maintainer:
    Email: emdemor415@gmail.com
    Copyright:
    Credits:
    License:
    Version:
    Status: in development

"""

import random,pygtc
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from tqdm       import tqdm
from .functions import distribute_among_walkers, riffle
from scipy      import stats
from os         import path
from numpy      import random


#__RESULTS_DIR__       =  path.join('cov19', '_results')
__ESTIMATE_OUT_FILE__ = 'sigle-parameter-estimates.csv'
__GTC_OUT_FILE__      = 'gtc-graphs.png'
__CRD_OUT_FILE__      = 'crd-curve.png'
__CURVES_PROJ_FILE__  = 'cases_projection.png'
__EP_PARAMETES_FILE__ = 'epidemiological_par.csv'
__TEND__              = 365

class stat_model:
    '''
    Description
    ----------
    The stat_model manipulates the global statistical parameters of the model

    Arguments
    ----------
    dataframe: country.dataframe
        A pandas dataframe with cov19 cases in the specific country

    ep_model:
        Details about the epidemiological model. Models mus bet implemented
        in the models.py module.

    par_labels: list (optional)
        A list of strings labeling the model parameters. If the user choose
        not to pass this, the code must treat it as ['p1',...,'pn'].

    par_est: numpy.array
        Raw estimates to the model parameter's values

    par_min: numpy.array (optional)
        Raw estimates to the model parameter's values

    par_max: numpy.array (optional)
        Raw estimates to the model parameter's values

    rescaling: float (optional)
        factor to reescaling the number of cases

    tend: float (optional)
        The maximum value of time. If the user choose not pass it, the code
        will get the last time from the dataset

    Parameters
    ----------

    self.dataframe: pandas.DataFrame
        Internal DataFrame object with data about covid cases.

    self.ep_model:
        Internal variable with the epidemiological model.

    '''

    def __init__(self,
                 dataframe,
                 ep_model,
                 par_est,
                 par_min=False,
                 par_max=False,
                 par_labels = False,
                 rescaling_by=1,
                 tend=False
                 ):

        print('[info]: Defining statistical model.')
        self.dataframe = dataframe
        self.ep_model = ep_model
        self.par_est = par_est
        self.rescale = rescaling_by
        self.n = 0

        if tend == False:
            self.tend = dataframe.days_list[-1]
        else:
            self.tend = tend

        if par_labels == False:
            self.par_labels = list(map(lambda a: 'p'+a,list('12345')))
        else:
            self.par_labels = par_labels

        self.mcmc_sample = False
        self.sample_imported = False
        self.ndim = len(par_est)




    def solve(self,parameters):

        """
        Call the solve method from epidemiological model object e returns it
        updated.

        Arguments
        ----------
        par: numpy.array
            Specific values for parameters

        x0: numpy.array
            Initial conditions for variables to be integrated

        tend: float
            Last value of time

        Parameters
        ----------
        void

        Returns
        -------
        ep_model

        """
        return self.ep_model(par = parameters, x0 = [10**(parameters[-1]),self.rescale,0,0],tend=self.tend)




    def chi_sqrd(self,
                 par,
                 fit_recovered = True,
                 fit_death = True,
                 fit_confirmed = True
                 ):

        """
        Evaluate the chi squared for a parametrical configuration

        Arguments
        ----------
        par: numpy.array
            Specific values for parameters

        fit_confirmed: boolean
            Controls of the chi squared will consider the confirmed cases

        fit_recovered : boolean
            Controls of the chi squared will consider the recovered cases

        fit_death: boolean
            Controls of the chi squared will consider the deaths cases


        Returns
        -------
        Chi2: float
            Value of chi squared

        """

        # model
        model = self.solve(par)
        t_model = model.days_list
        c_model = model.confirmed_list
        r_model = model.recovered_list
        d_model = model.death_list

        #dataset
        t_data = self.dataframe.days_list
        c_data = self.dataframe.confirmed_list
        r_data = self.dataframe.recovered_list
        d_data = self.dataframe.death_list

        # chi2 function
        chi2 = lambda model, data: (model-data)**2

        # stating chi2
        Chi2 = 0;

        # summing confirmed part
        if fit_confirmed:
            Chi2 += sum(list(map(chi2, list(np.interp(t_data, t_model, c_model)), c_data)))

        # summing recovered part
        if fit_recovered:
            Chi2 += sum(list(map(chi2, list(np.interp(t_data, t_model, r_model)), r_data)))

        # summing death part
        if fit_death:
            Chi2 += sum(list(map(chi2, list(np.interp(t_data, t_model, d_model)), d_data)))

        return Chi2




    def log_prob(self,
                 par,
                 fit_recovered = True,
                 fit_death = True,
                 fit_confirmed = True
                 ):
        """
        Evaluate the log o probability

        Arguments
        ----------
        par: numpy.array
            Specific values for parameters

        fit_confirmed: boolean
            Controls of the chi squared will consider the confirmed cases

        fit_recovered : boolean
            Controls of the chi squared will consider the recovered cases

        fit_death: boolean
            Controls of the chi squared will consider the deaths cases


        Returns
        -------
        ln: float
            Value of log P

        """

        lnP = - 0.5 * self.chi_sqrd(par,
                                    fit_recovered = fit_recovered,
                                    fit_death = fit_death,
                                    fit_confirmed = fit_confirmed
                                    )

        return lnP




    def metropolis_hastings(self,
                            n_points,
                            par_stp,
                            file_name = "mcmc.csv",
                            overwrite = False,
                            n_walkers = 1,
                            fit_recovered = True,
                            fit_death = True,
                            fit_confirmed = True
                            ):
        """
        Generate a MCMC sample through metropolis hasting algorithm and
        save it in a specified file

        Arguments
        ----------
        n_points: int
            Sample length

        par_stp: numpy.array
            The maximum displacement a walker can take at each step

        file_name: str
            A string containing the name of the file where the code will
            append values.

        overwrite: bool
            Variable passed by user controling if old files will be appended
            or overwritten

        n_walkers: int
            Number of random walkers

        fit_confirmed: boolean
            Controls of the chi squared will consider the confirmed cases

        fit_recovered : boolean
            Controls of the chi squared will consider the recovered cases

        fit_death: boolean
            Controls of the chi squared will consider the deaths cases

        Returns
        -------
        void

        """

        print('[info]: Generating a mcmc sample by metropolis-hastings algorithm.')

        # distributing the points to walkers
        n_walkers_list = distribute_among_walkers(n_points,n_walkers)

        # changing to tables directory
        #set_directory(tables_directory)

        # cleaning the old file
        if overwrite:
            open(file_name, 'w').close()

        # open a new file to append
        file = open(file_name, 'a')

        # writing the header
        if overwrite:
            file.write(''.join(riffle(self.par_labels,'\t'))+'\n')

        # looping through the walkers
        for ind_walker in range(len(n_walkers_list)):

            # evaluating the log-probability of the guess
            PAR = self.par_est
            log_PROB = self.log_prob(PAR,
                                    fit_recovered = fit_recovered,
                                    fit_death = fit_death,
                                    fit_confirmed = fit_confirmed)

            # looping through the walker amount of points
            for n in tqdm(range(n_walkers_list[ind_walker]),desc='[wlkr'+str(1+ind_walker)+']: '):

                # suggest new candidate from uniform distribution
                PAR_NEW  = list(map(lambda p,h: p + random.uniform(-1, 1)*h ,np.array(PAR),np.array(par_stp)))
                log_PROB_NEW = self.log_prob(PAR_NEW,
                                    fit_recovered = fit_recovered,
                                    fit_death = fit_death,
                                    fit_confirmed = fit_confirmed)

                # accept new candidate in Monte-Carlo fashing.
                if (log_PROB_NEW > log_PROB):
                    PAR  = PAR_NEW
                    log_PROB = log_PROB_NEW
                else:
                    u = random.uniform(0.0,1.0)
                    if (u < np.exp(log_PROB_NEW - log_PROB)):
                        PAR  = PAR_NEW
                        log_PROB = log_PROB_NEW

                # writing on file
                file.write(''.join(riffle(list(map(str,PAR)),'\t'))+'\n')

        file.close()

        # # updating mcmc_sample variable
        self.mcmc_sample = True




    def import_sample(self,
    	              file_name="mcmc.csv",
    				  filter_outliers = False
    				  ):

        """
        This method imports the file where the sample was saved

        Arguments
        ----------
        file_name: str
            A string containing the name of the file where the code will
            append values.

        Returns
        -------
        :pandas.DataFrame.info()
            Information about the sample imported
        """

        print('[info]: Reading mcmc sample file.')

        # setting tables directory
        #set_directory(tables_directory)

        #importing
        self.raw_sample_df = pd.read_csv(file_name,sep="\t")
        self.raw_sample = self.raw_sample_df.to_numpy()

        # Filtering outliers (WRONG)
        if filter_outliers:

            DIC = self.raw_sample_df.quantile(0.75)-self.raw_sample_df.quantile(0.25)
            beg = self.raw_sample_df.quantile(0.25)-1.75*DIC
            end = self.raw_sample_df.quantile(0.75)+1.75*DIC
            conditions = True
            for lab in self.par_labels:
            	conditions &= self.raw_sample_df[lab].between(beg[lab],end[lab])

            self.sample_df = self.raw_sample_df[conditions]
            self.sample = self.sample_df.to_numpy()
        else:
            self.sample_df = self.raw_sample_df
            self.sample = self.sample_df.to_numpy()


        # updating sample_imported
        self.sample_imported = True

        # number of data
        self.n = len(self.sample_df)


    def data_model_plot(self,par):

        """
        This method plots a graphics comparing the dataset with the curves
        related to the model evaluated with the parameters passed by user

        Arguments
        ----------
        par: numpy.array
            Array containig a specific parametric configuration

        Returns
        -------
        void

        """
        print('[info]: Ploting epidemiological model curves with dataset.')
        model = self.solve(par)

        # Confirmed Cases Plot
        plt.plot(model.days_list,model.confirmed_list,color='#27408B', label='Confirmed')
        plt.scatter(self.dataframe.days_list,self.dataframe.confirmed_list, color='#436EEE',alpha=0.9,s=5)

        # Recovered Cases Plot
        plt.plot(model.days_list,model.recovered_list,color='#008B45', label='Recovered')
        plt.scatter(self.dataframe.days_list,self.dataframe.recovered_list, color='#00CD66',alpha=0.9,s=5)

        # Death Cases Plot
        plt.plot(model.days_list,model.death_list,color='#EE7621', label='Deaths')
        plt.scatter(self.dataframe.days_list,self.dataframe.death_list, color='#FF7F24',alpha=0.9,s=5)


        plt.legend(loc="upper left")
        plt.xlabel('days after first case')
        plt.ylabel('thousands of people')
        plt.grid()
        plt.savefig(__CRD_OUT_FILE__)
        plt.show()




    def single_parameter_estimates(self,alpha=0.3173,est_outfile = __ESTIMATE_OUT_FILE__):

        """
        Returns the interval with confidence alpha.

        Arguments
        ----------
        alpha: float
            Confidence level of interval estimate

        Returns
        -------
        void

        """

        # check if user imported a sample
        if not self.sample_imported:
            print('[error]: you must read a mcmc sample file first.')
            return []

        else:
            print('[info]: Evaluating statistical properties of sample.')
            var = 100*alpha/2

            med = self.raw_sample_df.mean()

            interval = np.array(list(map(lambda index: np.percentile(self.raw_sample[:, index], [50,var, 100-var]),list(range(self.ndim)))))
            df_results = pd.DataFrame({ 'Parameters':self.par_labels,
                                        'Mean': med,
                                        'Medians':interval[:,0],
                                        '1 sig interval min': interval[:,1],
                                        '1 sig interval max': interval[:,2]
                                      })

            df_results .to_csv(est_outfile,sep='\t',index=False)
            #var = list(map(lambda label, estim: label+str(estim[0]))
            self.estimates = df_results
            return df_results




    def gtc_plot(self,
                 truths=None,
                 n_contour_levels=2,
                 figure_size = 8,
                 custom_label_font = {'family':'DejaVu Sans', 'size':10},
                 custom_tick_font = {'family':'DejaVu Sans', 'size':8},
                 save_figure = False,
                 file_name = __GTC_OUT_FILE__,
                 show = True
            ):

        """
        This method plots graphics for each 2-dimensional marginal
        probability density and the 1-dimensional marginal probability
        related to the parameters

        Arguments
        ----------
        ...
        ...


        Returns
        -------
        ...
        ...

        """
        plt.close()
        GTC = pygtc.plotGTC(chains=[self.sample ],
                    truths = truths,
                    paramNames = self.par_labels,
                    nContourLevels = n_contour_levels,
                    figureSize = figure_size,
                    customLabelFont = custom_label_font,
                    customTickFont = custom_tick_font
                   )

        if save_figure:
            GTC.savefig(file_name)

        if show: plt.show()
        plt.close()
        #return GTC




    def evaluate_epidemiological_parameters(self,
                                            tend = __TEND__,
                                            overwrite = True,
                                            file_name = __EP_PARAMETES_FILE__,
                                            sample = 0.1
                                            ):

        """
        This method gives parametric estimates to parameters

        Arguments
        ----------
        ...
        ...

        Returns
        -------
        ...
        ...

        """

        print('[info]: Generating confidence regions for epidemiological curves.')

        # checking if user want all sample or just a part
        if (type(sample) == int) and (0 < sample < len(self.sample)):
            sample_list = [self.sample[index] for index in list(random.choice(range(len(self.sample)),sample))]

        elif (type(sample) == float) and (0. < sample <= 1.0):
            sample = int(sample*round(len(self.sample)))
            sample_list = [self.sample[index] for index in list(random.choice(range(len(self.sample)),sample))]
        else:
            sample_list = self.sample

        # updating sample_imported
        if not self.sample_imported:
            print('[error]: You need to generate and import a sample.')

        else:

            # changing to tables folder
            #set_directory(tables_directory)

            # cleaning the old file
            if overwrite:
                open(file_name, 'w').close()

            # open a new file to append
            file = open(file_name, 'a')

            # writing the header
            if overwrite:
                PAR_LAB = self.par_labels + ['Confirmed','Deaths']
                file.write(''.join(riffle(PAR_LAB,'\t'))+'\n')

            # loop through the sample parameters
            for par in tqdm(sample_list,desc='[wlkr1]: '):

                # initial conditions
                x0 = [10**(par[-1]),self.rescale,0,0]

                #solving equations
                msird_av = self.ep_model(par, x0,tend)
                msird_av.solve()

                # time list
                t = msird_av.days_list

                # Confirmed Cases Plot
                c_model = msird_av.confirmed_list

                # Recovered Cases Plot
                r_model = msird_av.recovered_list

                # Death Cases Plot
                d_model = msird_av.death_list

                PAR = np.concatenate((par, np.array([np.interp(tend, t, c_model),np.interp(tend, t, d_model)])))

                # writing on file
                file.write(''.join(riffle(list(map(str,PAR)),'\t'))+'\n')

            file.close()




    def plot_curves(self,
    				par_est,
                    tend  = __TEND__,
                    alpha = 0.02,
                    show = True,
                    save_figure = True,
                    file_name = __CURVES_PROJ_FILE__,
                    plot_title = ''
                    ):

        """
        This method gives parametric estimates to parameters

        Arguments
        ----------
        ...
        ...

        Returns
        -------
        ...
        ...

        """

        print('[info]: Plotting curves.')
        plt.close()

        # Solving Equations
        x0 = [10**(par_est[-1]),self.rescale,0,0]
        model = self.ep_model(par_est, x0,tend)
        model.solve()

        # Time list
        t = model.days_list

        #Confirmed Cases Plot
        c_model = model.confirmed_list
        plt.plot(t,c_model,color='#27408B', label='Confirmed',zorder=10)
        plt.scatter(self.dataframe.days_list,
                    self.dataframe.confirmed_list,
                    color='#03254c',
                    s=9,
                    zorder=13)

        # Recovered Cases Plot
        r_model = model.recovered_list
        plt.plot(t,r_model,color='#008B45', label='Recovered',zorder=10)
        plt.scatter(self.dataframe.days_list,
                    self.dataframe.recovered_list,
                    color='#1e663b',
                    alpha=1.,
                    s=9,
                    zorder=12)

        # Death Cases Plot
        d_model = model.death_list
        plt.plot(t,d_model,color='#EE7621', label='Deaths',zorder=10)
        plt.scatter(self.dataframe.days_list,self.dataframe.death_list, color='#FF7F24',alpha=1.,s=9,zorder=11)

        #plt.show()
        final_deaths = []
        final_cases  = []

        for index in tqdm(list(random.choice(range(len(self.sample)),500)),desc='[wlkr1]: '):

            par_rdn = self.sample[index]
            # Solving Equations
            x0 = [10**(par_rdn[-1]),self.rescale,0,0]
            model = self.ep_model(par_rdn, x0,tend)
            model.solve()
            t = model.days_list

            # Confirmed Cases Plot
            c_model = model.confirmed_list
            plt.plot(t,c_model,color='#87CEFA',  alpha=alpha, linewidth=3)

            # Recovered Cases Plot
            r_model = model.recovered_list
            plt.plot(t,r_model,color='#98e0b5',alpha=alpha, linewidth=3)

            # Death Cases Plot
            d_model = model.death_list
            plt.plot(t,d_model,color='#FFDAB9', alpha=alpha, linewidth=3)

        plt.title(plot_title)
        plt.legend(loc="upper left")
        plt.xlabel('days after first case')
        plt.ylabel('number of people [✕'+str(int(1/self.rescale))+']')
        plt.grid()
        if save_figure:
            plt.savefig(file_name)
        if show: plt.show()
        plt.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

Description
----------
This module gives general functions
----------
    Author: Eduardo M.  de Morais
    Maintainer:
    Email: emdemor415@gmail.com
    Copyright:
    Credits:
    License:
    Version:
    Status: in development
    
"""

import os, pygtc, configparser, json, sys
from os import path


# getting notebook directory
notebook_directory = os.getcwd()


    
def import_parameters(file_name: str):
    """
    Description
    ----------
    This function returns the directory address of notebook 
    and the tables directories.
    """

    # reading config file
    if path.exists(file_name):
        config = configparser.ConfigParser()
        config.read(file_name)

        #import parameters
        result = {
                  'ep_model'         : json.loads(config.get("MODEL", "ep_model")),
                  'scl_factor'       : json.loads(config.get("MODEL", "scl_factor")),
                  'par_labels'       : json.loads(config.get("MODEL", "par_labels")),
                  'par_dim'          : len(json.loads(config.get("MODEL", "par_labels"))),
                  'par_est'          : json.loads(config.get("MODEL", "par_est")),
                  'par_stp'          : json.loads(config.get("MODEL", "par_stp")),

                  'data_filename'    : json.loads(config.get("DATASET", "data_filename")),
                  'data_url'         : json.loads(config.get("DATASET", "data_url")),
                  'data_update'      : json.loads(config.get("DATASET", "update")),
                  'last_index'      : json.loads(config.get("DATASET", "last_index")),

                  'country_1'        : json.loads(config.get("COUNTRY", "country_1")),

                  'mcmc_file_name'   : json.loads(config.get("MCMC", "file_name")),
                  'generate_mcmc'    : json.loads(config.get("MCMC", "generate_mcmc")),
                  'overwrite_file'   : json.loads(config.get("MCMC", "overwrite_file")),
                  'sample_length'    : json.loads(config.get("MCMC", "sample_length")),
                  'n_walkers'        : json.loads(config.get("MCMC", "n_walkers")),
                  'fit_confirmed'    : json.loads(config.get("MCMC", "fit_confirmed")),
                  'fit_death'        : json.loads(config.get("MCMC", "fit_death")),
                  'fit_recovered'    : json.loads(config.get("MCMC", "fit_recovered")),
                  'gtc_plot'         : json.loads(config.get("MCMC", "gtc_plot")),
                  'gtc_out_file'     : json.loads(config.get("MCMC", "gtc_out_file")),

                  'ep_file_name'      : json.loads(config.get("EPID", "ep_file_name")),
                  'ep_par_prop'       : json.loads(config.get("EPID", "ep_par_prop")),
                  'ep_par_sample'     : json.loads(config.get("EPID", "ep_par_sample")),
                  'ep_par_overwrite'  : json.loads(config.get("EPID", "ep_par_overwrite")),
                  'ep_plot_curves'    : json.loads(config.get("EPID", "ep_plot_curves")),
                  'ep_est_out_file'   : json.loads(config.get("EPID", "ep_est_out_file")),
                  'ep_crd_out_file'   : json.loads(config.get("EPID", "ep_crd_out_file")),
                  'ep_crv_prj_file'   : json.loads(config.get("EPID", "ep_crv_prj_file")),
                  'ep_paramet_file'   : json.loads(config.get("EPID", "ep_paramet_file")),
                  'ep_t_end'         : json.loads(config.get("EPID", "ep_t_end"))
                  }

    else:
        result = {}

    return result



    
def set_dir_struct():
    """
    Description
    ----------
    This function returns the directory address of notebook 
    and the tables directories.
    """
    
    lista=[];
    lista.append(notebook_directory);
    lista.append(os.path.join("cov19","_tables"));
    return lista




def file_names(direct = os.getcwd()):
    """
    Description
    ----------
    This function receive a directory address as argument
    and return a list of strings with the names of all files
    inside it.
    """
    print(direct)
    return os.listdir(direct)




def riffle(lst, item):
    """
    Description
    ----------
    Put the item between each element of the list "lst".
    """
    result = [item] * (len(lst) * 2 - 1)
    result[0::2] = lst
    return result




def set_directory(direc):
    """
    Description
    ----------
    This function receive a directory addres as argument
    and set this as the current work directory.
    """
    os.chdir(direc)
    
    
    
    
def distribute_among_walkers(quantity,walkers):
    
    '''
    Description
    ----------
    This function returns a list with size equal to "walkers".
    Its divide the "quantity" for the number of walkers keeping
    the modules on the firsts.
    '''
    
    div_int = quantity//walkers
    div_mod = quantity%walkers

    walkers_list = []
    
    for index in range(walkers):
        mod_sum = 1 if index < div_mod else 0
        walker_stps = div_int + mod_sum
        walkers_list.append(walker_stps)
        
    return(walkers_list)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Description
----------
This module gives functions and classes to manipulate epidemiolgical models

Informations
----------
    Author: Eduardo M.  de Morais
    Maintainer:
    Email: emdemor415@gmail.com
    Copyright:
    Credits:
    License:
    Version:
    Status: in development
    
"""

import numpy as np
from scipy.integrate import odeint


class mod_sird:
    '''
    Description
    ----------
    This class defines a epidemiological model to explained the total, 
    recovered, active infected and deaths cases of the COVID pandemic desease 
    in a specific country.
    
    Its equations descrive the relations between the 4 functions s,i,r and d 
    describing respectively the number of susceptible, active infected, 
    recovered and dead individual
    
    The diference bewtween Modified SIRD and the usual SIRD model is in the 
    presence of an aditional parameters alpha as a power on i(t) in 
    differential equations 
    
    
    Arguments
    ----------
        par: numpy.array
            Specific values for parameters
            
        x0: numpy.array
            Initial conditions for variables to be integrated
            
        tend: float
            Last value of time
            
        tbeg: float (optional)
            First value of time
            
        tend: float
            Last value of time
            
        npoints: int
            Number of interpolated points evaluated by model
            
        
    Parameters
    ----------
        self.par: numpy.array
            Specific values for parameters

        self.x0: numpy.array
            Initial conditions for variables to be integrated

        self.tbeg: float (optional)
            First value of time 

        self.tend:  float
            Last value of time

        self.npoints: int
            Number of interpolated points evaluated by model

        self.x: numpy.array
            Array with 4 columns and n lines containing the susceptible,
            active infected, recovered and death functions evaluated in
            time.
        
        self.days_list:
            Array with  n lines containing the time values
        
        self.confirmed_list:
            Array with  n lines containing the evaluated confirmed cases

        self.recovered_list:
            Array with  n lines containing the evaluated recovered cases

        self.death_list:
            Array with  n lines containing the evaluated deaths cases
        
    '''
    
    def __init__(self,
                 par,
                 x0,
                 tend,
                 tbeg=0,
                 npoints=100
                 ):
        
        self.par = par
        self.x0 = x0
        self.tbeg = tbeg
        self.tend = tend
        self.npoints = npoints
        self.name = "MSIRD"
        self.solve()




    def diff_eq(self,x,t,par):
        
        """
        Function resturning the differential equations of the model

        Arguments
        ----------
        x: numpy.array
            Array with 4 columns and n lines containing the susceptible,
            active infected, recovered and death functions evaluated in
            time.
            
        t: float
            Time value where the equations will be avaluated
            
        par: numpy.array
            Specific values for parameters
            
        Parameters
        ----------
        void
        
        Returns
        -------
        : list
            Differential equations evaluated for the parameters an time
            passed by user
        
        """
        
        # setting the functions
        s = x[0]
        i = x[1]
        r = x[2]
        d = x[3]
    
        # mathematical equations
        DiffS = -par[1]*(s/10**par[4])*(i**par[0])
        DiffI = (par[1]*(s/10**par[4]) - par[2] - par[3])*(i**par[0])
        DiffR = par[2]*(i**par[0])
        DiffD = par[3]*(i**par[0])
        
        return [DiffS,DiffI,DiffR,DiffD]
    



    def solve(self,
              notifications = False
              ):
        
        """
        This method imports the file where the sample was saved
        
        Arguments
        ----------
        file_name: str
            A string containing the name of the file where the code will
            append values.
            
        Returns
        -------
        :pandas.DataFrame.info()
            Information about the sample imported
        """

        if notifications:
            print('[info]: Solving differential equations for '+self.name+' model. ')
            

        
        # getting the time values
        self.days_list = np.linspace(self.tbeg,self.tend,self.npoints)

        # calling the odeint method to solve the diff. equations
        self.x = odeint(self.diff_eq,self.x0,self.days_list,args = (self.par,))
        '''
        Its important to note that (par_est,) is the way to define a tuple
        with just one element. When we put (par_est), the  parenteses won't
        indicate a tuple
        '''
        
        #setting the variables
        self.confirmed_list = self.x[:,1] + self.x[:,2] + self.x[:,3]
        self.recovered_list = self.x[:,2]
        self.death_list = self.x[:,3]
    
            

class gmsird:
    '''
    Description
    ----------

    Arguments
    ----------
        par: numpy.array
            Specific values for parameters
            
        x0: numpy.array
            Initial conditions for variables to be integrated
            
        tend: float
            Last value of time
            
        tbeg: float (optional)
            First value of time
            
        tend: float
            Last value of time
            
        npoints: int
            Number of interpolated points evaluated by model
            
        
    Parameters
    ----------
        self.par: numpy.array
            Specific values for parameters

        self.x0: numpy.array
            Initial conditions for variables to be integrated

        self.tbeg: float (optional)
            First value of time 

        self.tend:  float
            Last value of time

        self.npoints: int
            Number of interpolated points evaluated by model

        self.x: numpy.array
            Array with 4 columns and n lines containing the susceptible,
            active infected, recovered and death functions evaluated in
            time.
        
        self.days_list:
            Array with  n lines containing the time values
        
        self.confirmed_list:
            Array with  n lines containing the evaluated confirmed cases

        self.recovered_list:
            Array with  n lines containing the evaluated recovered cases

        self.death_list:
            Array with  n lines containing the evaluated deaths cases
        
    '''
    
    def __init__(self,
                 par,
                 x0,
                 tend,
                 tbeg=0,
                 npoints=100
                 ):
        
        self.par = par
        self.x0 = x0
        self.tbeg = tbeg
        self.tend = tend
        self.npoints = npoints
        self.name = "MSIRD"
        self.solve()




    def diff_eq(self,x,t,par):
        
        """
        Function resturning the differential equations of the model

        Arguments
        ----------
        x: numpy.array
            Array with 4 columns and n lines containing the susceptible,
            active infected, recovered and death functions evaluated in
            time.
            
        t: float
            Time value where the equations will be avaluated
            
        par: numpy.array
            Specific values for parameters
            
        Parameters
        ----------
        void
        
        Returns
        -------
        : list
            Differential equations evaluated for the parameters an time
            passed by user
        
        """
        
        # setting the functions
        s, i, r, d = x

        n = len(par)-1
        # a_s = 0.85
        a_s, v, b, c, d = par[:n]
        v = v/1000
        a_r = a_s
        a_d = a_s
        s0 = 10**par[-1]
    
        # mathematical equations
        DiffS = -b*(s/s0)*(i**a_s)
        DiffI = b*(s/s0)*(i**a_s) - c*(i**(a_s+v*t)) - d*(i**(a_s+v*t))
        DiffR = c*(i**(a_s+v*t))
        DiffD = d*(i**(a_s+v*t))
        
        return [DiffS,DiffI,DiffR,DiffD]
    



    def solve(self,
              notifications = False
              ):
        
        """
        This method imports the file where the sample was saved
        
        Arguments
        ----------
        file_name: str
            A string containing the name of the file where the code will
            append values.
            
        Returns
        -------
        :pandas.DataFrame.info()
            Information about the sample imported
        """

        if notifications:
            print('[info]: Solving differential equations for '+self.name+' model. ')
            

        
        # getting the time values
        self.days_list = np.linspace(self.tbeg,self.tend,self.npoints)

        # calling the odeint method to solve the diff. equations
        self.x = odeint(self.diff_eq,self.x0,self.days_list,args = (self.par,))
        '''
        Its important to note that (par_est,) is the way to define a tuple
        with just one element. When we put (par_est), the  parenteses won't
        indicate a tuple
        '''
        
        #setting the variables
        self.confirmed_list = self.x[:,1] + self.x[:,2] + self.x[:,3]
        self.recovered_list = self.x[:,2]
        self.death_list = self.x[:,3]
    

class gen_sird:
    '''
    Description
    ----------
    
    Arguments
    ----------
        par: numpy.array
            Specific values for parameters
            
        x0: numpy.array
            Initial conditions for variables to be integrated
            
        tend: float
            Last value of time
            
        tbeg: float (optional)
            First value of time
            
        tend: float
            Last value of time
            
        npoints: int
            Number of interpolated points evaluated by model
            
        
    Parameters
    ----------
        self.par: numpy.array
            Specific values for parameters

        self.x0: numpy.array
            Initial conditions for variables to be integrated

        self.tbeg: float (optional)
            First value of time 

        self.tend:  float
            Last value of time

        self.npoints: int
            Number of interpolated points evaluated by model

        self.x: numpy.array
            Array with 4 columns and n lines containing the susceptible,
            active infected, recovered and death functions evaluated in
            time.
        
        self.days_list:
            Array with  n lines containing the time values
        
        self.confirmed_list:
            Array with  n lines containing the evaluated confirmed cases

        self.recovered_list:
            Array with  n lines containing the evaluated recovered cases

        self.death_list:
            Array with  n lines containing the evaluated deaths cases
        
    '''
    
    def __init__(self,par,x0,tend,tbeg=0,npoints=100):
        
        self.par = par
        self.x0 = x0
        self.tbeg = tbeg
        self.tend = tend
        self.npoints = npoints
        self.solve()




    def diff_eq(self,x,t,par):
        
        """
        Function resturning the differential equations of the model

        Arguments
        ----------
        x: numpy.array
            Array with 4 columns and n lines containing the susceptible,
            active infected, recovered and death functions evaluated in
            time.
            
        t: float
            Time value where the equations will be avaluated
            
        par: numpy.array
            Specific values for parameters
            
        Parameters
        ----------
        void
        
        Returns
        -------
        : list
            Differential equations evaluated for the parameters an time
            passed by user
        
        """
        
        # setting the functions
        s = x[0]
        i = x[1]
        r = x[2]
        d = x[3]

        n = len(par)-1
        # a_s = 0.85
        a_s, a_r, a_d, b, c, d = par[:n]

        s0 = 10**par[-1]
    
        # mathematical equations
        DiffS = -b*(s/s0)*(i**a_s)
        DiffI = b*(s/s0)*(i**a_s) - c*(i**a_r) - d*(i**a_d)
        DiffR = c*(i**a_r)
        DiffD = d*(i**a_d)
        
        return [DiffS,DiffI,DiffR,DiffD]
    



    def solve(self):
        
        """
        This method imports the file where the sample was saved
        
        Arguments
        ----------
        file_name: str
            A string containing the name of the file where the code will
            append values.
            
        Returns
        -------
        :pandas.DataFrame.info()
            Information about the sample imported
        """
        
        # getting the time values
        self.days_list = np.linspace(self.tbeg,self.tend,self.npoints)

        # calling the odeint method to solve the diff. equations
        self.x = odeint(self.diff_eq,self.x0,self.days_list,args = (self.par,))
        '''
        Its important to note that (par_est,) is the way to define a tuple
        with just ode element. When we put (par_est), the  parenteses won't
        indicate a typle
        '''
        
        #setting the variables
        self.confirmed_list = self.x[:,1] + self.x[:,2] + self.x[:,3]
        self.recovered_list = self.x[:,2]
        self.death_list = self.x[:,3]
    
            


from setuptools import setup

with open("README.md", "r") as fh:
    readme = fh.read()

setup(name='cov19',
    version='0.0.3',
    url='https://github.com/emdemor/cov19',
    license='MIT License',
    author='Eduardo M. de Morais',
    long_description=readme,
    long_description_content_type="text/markdown",
    author_email='emdemor415@gmail.com',
    keywords='covid-19',
    description=u'Prediction for the spread of COVID-19 using Epidemiological Models',
    packages=['cov19'],
    install_requires=['numpy','scipy','pandas','pygtc','setuptools','tqdm'],)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

Description
----------
The region object contains information and actions 
related the regions that are important to understand 
the Covid dissemination

Informations
----------
    Author: Eduardo M.  de Morais
    Maintainer:
    Email: emdemor415@gmail.com
    Copyright:
    Credits:
    License:
    Version:
    Status: in development
    
"""



class region:
    '''
    Description
    ----------
    The region object contains information and actions related the regions
    that are important to understand the Covid dissemination

    Arguments
    ----------
    name: str
        String passed by the user to label the country
        
    population: int
        Integer number passed by the user counting people in the related 
        country

    Parameters
    ----------
    self.name: str
        String used to label the country.
        
    self.population: int
        Integer number with the number of people in the related country
        
    self.dataframe: pandas.DataFrame
        DataFrame object with data about covid cases.
        
    self.days_list: numpy.ndarray
        Numpy array with the days after first confirmed covid case
        
    self.confirmed_list: numpy.ndarray
        Numpy array with the number of confirmed cases. Must have the same 
        length of self.days_list
        
    self.recovered_list: numpy.ndarray
        Numpy array with the number of recovered cases. Must have the same 
        length of self.days_list
        
    self.death_list: numpy.ndarray
        Numpy array with the number of deaths cases. Must have the same 
        length of self.days_list
        
    '''
    
    def __init__(self,
        name,
        population,
        df,
        rescaling_by  = 1,
        last_index    = -1,
        confirmed_col = 'Confirmed',
        recovered_col = 'Recovered',
        deaths_col    = 'Deaths',
        actives_col   = 'Actives'
        ):
        
        """
        Receive a global dataframe with COVID data from microsoft repository,
        select those related to informed country and sets the country class' 
        parameters

        Parameters
        ----------
        df_Country : pandas dataframe
            Filter the information related to the country
            
        first_case : timestamp
            Date of fist covid case
        
        Returns
        -------
        void
        
        """
        
        # getting the data from country
        df_Region = df.loc[(df[reg_column] == self.name) & (df['AdminRegion1'] == '')]
        
        # getting the first case
        first_case = df_Region['Updated'].min()
        
        # couting teh days after first case
        df_Region['Days'] = (df_Region['Updated'] - first_case).dt.days.astype(float)
        
        # rescaling cases by a factor
        df_Region[['Confirmed','Actives','Deaths','Recovered']] = df_Region[['Confirmed','Actives','Deaths','Recovered']]*rescaling_by
        self.dataframe = df_Region
        
        # converting columns to lists
        self.days_list = df_Region['Days'].to_numpy()
        self.confirmed_list = df_Region['Confirmed'].to_numpy()
        self.death_list = df_Region['Deaths'].to_numpy()
        self.recovered_list = df_Region['Recovered'].to_numpy()
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Description
----------
This is the main file of aplication

Informations
----------
    Author: Eduardo M.  de Morais
    Maintainer:
    Email: emdemor415@gmail.com
    Copyright:
    Credits:
    License:
    Version:
    Status: in development
"""

from cov19 import read_dataset,country,mod_sird,gmsird,gen_sird
from cov19.stat import *
from cov19.functions import *
from numpy import random

import numpy as np
import pygtc, configparser, json, sys
import matplotlib.pyplot as plt
import cov19.dataset as ds

def main(file_name):

	# import parameter from configuration file
	param = import_parameters(file_name)

	# reading dataset
	df = read_dataset(update_data = param['data_update'],
                	  url         = param['data_url'] ,
                      local_dataset_filename = param['data_filename'])

	# getting country information
	brasil_df = country(name       = param['country_1'][0],
		                population = param['country_1'][1]
		                )
	brasil_df.dataset(df,
		rescaling_by=1/param['scl_factor'],
		last_index = param['last_index'])

	
	# If user chooses 'mod_sird'
	if param['ep_model'] == 'Mod-SIRD':

		# defining statistical model
		StatModel = stat_model(brasil_df,
		                       mod_sird,
		                       param['par_est'],
		                       rescaling_by = 1/param['scl_factor'],
		                       par_labels   = param['par_labels']
		                       )

	# If user chooses 'gmsird'
	elif param['ep_model'] == 'GM-SIRD':

		# defining statistical model
		StatModel = stat_model(brasil_df,
		                       gmsird,
		                       param['par_est'],
		                       rescaling_by = 1/param['scl_factor'],
		                       par_labels   = param['par_labels']
		                       )


	# If user chooses 'gen_sird'
	elif param['ep_model'] == 'Gen-SIRD':

		# defining statistical model
		StatModel = stat_model(brasil_df,
		                       gen_sird,
		                       param['par_est'],
		                       rescaling_by = 1/param['scl_factor'],
		                       par_labels   = param['par_labels']
		                       )


	# generating a mcmc sample by metropolis-hastings algorithm
	if param['generate_mcmc']:
		StatModel.metropolis_hastings(n_points  = param['sample_length'],
		                              par_stp   = param['par_stp'],
		                              overwrite = param['overwrite_file'],
		                              file_name = param['mcmc_file_name']
		                              )

	# import mcmc sample
	StatModel.import_sample(file_name = param['mcmc_file_name'],
                        	filter_outliers = True)

	# evaluating estimates for parameters
	SingleParameterEstimates = StatModel.single_parameter_estimates(alpha=0.3173,est_outfile = param["ep_est_out_file"])


	if param['ep_par_prop']:
			StatModel.evaluate_epidemiological_parameters(
									overwrite = param['ep_par_overwrite'],
									file_name = param['ep_file_name'],
									sample    = param['ep_par_sample']
	                                )


	# generating GTC plot
	if param['gtc_plot']:
		StatModel.gtc_plot(save_figure=True,show=False,file_name = param['gtc_out_file'])

	# plotting curves
	if param['ep_plot_curves']:
		StatModel.plot_curves(SingleParameterEstimates['Medians'].to_list(),
	                    save_figure = True,
	                    show        = False,
	                    file_name = param['ep_crv_prj_file'],
                        plot_title = param['country_1'][0]#+' , '+param['ep_model']
	                    )


if __name__ == "__main__":

	args = sys.argv;

	if (len(sys.argv) != 2):
		print('[Error]: You must pass a parameter file to '+str(args[0]))
	else:
		main(args[1])
	print('[status]: Finished.')
